/*
 * fwho - display login records
 *
 * synopsis:
 *	fwho [+format] [-C] [-l lastlog ..] [-u utmp ..] [-w wtmp ..]
 *
 * description:
 *	fwho displays lastlog(5), utmp(5), wtmp(5) format data and
 *	timestamps of those files and the time the user logged in with
 *	endian changing capabilities for the time data. The files may
 *	be compressed by gzip(1).
 *
 * options:
 *	+format	specify strftime(3) format, default "%Y/%m/%d-%H:%M:%S".
 *	-C	specify changing endian for the time data.
 *	-l	specify lastlog format files.
 *	-u	specify utmp format files.
 *	-w	specify wtmp format files.
 *
 * Copyright (c) 2009 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>	/* lstat, etc */
#include <time.h>	/* localtime, etc */
#include <utmp.h>	/* lastlog, utmp, wtmp, etc */
#include <sys/param.h>	/* MAXPATHLEN, etc */
#if defined(DISABLE_ZLIB)
typedef FILE *File;
#else
#include <zlib.h>	/* gzopen, etc */
typedef gzFile File;
#define fopen gzopen
#define fread(p,s,n,f) gzread(f,p,(s)*(n))
#define feof gzeof
#define fclose gzclose
#endif

/* not completely the same as BSD strlcpy */
#define strlcpy(d, s, z) do {\
  strncpy(d, s, z);\
  if ((z) > 0)\
    d[(z)-1] = '\0';\
} while (0)

/* swap for arbitrary type */
#define swap(a,b) do {\
  struct tmp_t { char t[sizeof(a)]; } tmp;\
  tmp = *(struct tmp_t *)&(a);\
  *(struct tmp_t *)&(a) = *(struct tmp_t *)&(b);\
  *(struct tmp_t *)&(b) = tmp;\
} while(0)

/* change endian for arbitrary type */
#define chend(p) do {\
  int i;\
  for (i=0; i<sizeof(*(p))/2; i++)\
    swap(((char *)p)[i], ((char *)p)[sizeof(*(p))-i-1] );\
} while (0)

static char ftm[256] = "%Y/%m/%d-%H:%M:%S";
static int sw_chend = 0;

int usage(int rc)
{
  puts("usage: fwho [+format] [-C] [-l lastlog ..] [-u utmp ..] [-w wtmp ..]");
  return rc;
}

int fprint_ftimestamp(const char *path, FILE *ofp)
{
  struct stat sb;
  time_t *ti[] = { &sb.st_ctime, &sb.st_mtime, /*&sb.st_atime,*/ };
  struct tm tm[sizeof(ti)/sizeof(ti[0])];
  size_t i;
  char buf[256];

  if (lstat(path, &sb)) return 1;
  fprintf(ofp, "%s: ", path);
  for (i=0; i<sizeof(tm)/sizeof(tm[0]); i++) {
    localtime_r(ti[i], &tm[i]);
    strftime(buf, sizeof(buf), ftm, &tm[i]);
    fprintf(ofp, "%s", buf);
    fprintf(ofp, "%s", (i != sizeof(tm)/sizeof(tm[0])-1) ? " - " : "\n");
  }
  return 0;
}

int fprint_lastlog(const char *path, FILE *ofp)
{
  File ifp;
  size_t z, i = 0, c = 0;
  struct lastlog u;
  struct tm tm;
  char buf[256], line[UT_LINESIZE+1], host[UT_HOSTSIZE+1];

  if (!strcmp(path, "-")) ifp = stdin;
  else if (!(ifp = fopen(path, "r"))) return 1;
  else fprint_ftimestamp(path, ofp);
  while ((z=fread(&u, sizeof(u), 1, ifp))) {
    if (u.ll_time) {
      if (sw_chend) chend(&u.ll_time);
      strftime(buf, sizeof(buf), ftm, localtime_r(&u.ll_time, &tm));
      strlcpy(line, u.ll_line, sizeof(line));
      strlcpy(host, u.ll_host, sizeof(host));
      fprintf(ofp, "%s\t%lu\t%-8s\t%-16s\n", buf, i, line, host);
      c++;
    }
    i++;
    if (feof(ifp)) break;
  }
  fprintf(ofp, "entry/record: %lu/%lu\n\n", c, i);
  if ((FILE *)ifp != stdin)
    fclose(ifp);
  return 0;
}
int fprint_utmp(const char *path, FILE *ofp)
{
  File ifp;
  size_t z, i = 0, c = 0;
  struct utmp u;
  struct tm tm;
  char buf[256], line[UT_LINESIZE+1], name[UT_NAMESIZE+1], host[UT_HOSTSIZE+1];

  if (!strcmp(path, "-")) ifp = stdin;
  else if (!(ifp = fopen(path, "r"))) return 1;
  else fprint_ftimestamp(path, ofp);
  while ((z=fread(&u, sizeof(u), 1, ifp))) {
    if (u.ut_time) {
      if (sw_chend) chend(&u.ut_time);
      strftime(buf, sizeof(buf), ftm, localtime_r(&u.ut_time, &tm));
      strlcpy(line, u.ut_line, sizeof(line));
      strlcpy(name, u.ut_name, sizeof(name));
      strlcpy(host, u.ut_host, sizeof(host));
      fprintf(ofp, "%s\t%-8s\t%-8s\t%-16s\n", buf, line, name, host);
      c++;
    }
    i++;
    if (feof(ifp)) break;
  }
  fprintf(ofp, "entry/record: %lu/%lu\n\n", c, i);
  if ((FILE *)ifp != stdin)
    fclose(ifp);
  return 0;
}
int fprint_wtmp(const char *path, FILE *ofp)
{
  return fprint_utmp(path, ofp);
}

int main(int argc, char *argv[])
{
  int a, i = 0;

  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--help")==0) return usage(0);
    else if (argv[a][0] == '+') strlcpy(ftm, &argv[a][1], sizeof(ftm));
    else if (!strcmp(argv[a], "-C")) sw_chend = !0;
    else if (!strcmp(argv[a], "-l"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_lastlog(argv[i], stdout);
    else if (!strcmp(argv[a], "-u"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_utmp(argv[i], stdout);
    else if (!strcmp(argv[a], "-w"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_wtmp(argv[i], stdout);

  if (!i) {
    char wtmp[][MAXPATHLEN] = {
      _PATH_WTMP".4.gz",
      _PATH_WTMP".3.gz",
      _PATH_WTMP".2.gz",
      _PATH_WTMP".1.gz",
      _PATH_WTMP".0.gz",
      _PATH_WTMP,
    };

    fprint_lastlog(_PATH_LASTLOG, stdout);
    fprint_utmp(_PATH_UTMP, stdout);
    for (i=0; i<sizeof(wtmp)/sizeof(wtmp[0]); i++)
      fprint_wtmp(wtmp[i], stdout);
  }
  return 0;
}
